import lxml.etree as ET
import logging
import numpy as np
from . import user_namespace as un
from . import utilities as util
from . import group
from . import label

log = logging.getLogger('prefigure')

# Add a graphical element describing a Riemann sum
# left, right, midpoint, trapezoid, simpsons, upper, lower, samples
def riemann_sum(element, diagram, parent, outline_status):
    diagram.add_id(element, element.get('id'))
    element_id = element.get('id')

    # Author may give a partition and samples so we'll grab them first
    partition = None
    samples = None
    if element.get('partition', None) is not None:
        partition = un.valid_eval(element.get('partition'))
        N = len(partition) - 1
    if element.get('samples', None) is not None:
        samples = un.valid_eval(element.get('samples'))

    # If a parition is not defined, we'll create one
    if partition is None:
        bbox = diagram.bbox()
        domain = element.get('domain')
        if domain == None:
            domain = [bbox[0], bbox[2]]
        else:
            domain = un.valid_eval(domain)
        try:
            N = int(element.get('N'))
        except:
            log.error(f"Error in <riemann-sum> setting N={element.get('N')}")
            return
        partition = np.linspace(domain[0], domain[1], N+1)

    # some rules will be constant on each interval
    constant_rules = {'left', 'right', 'midpoint', 'user-defined', 'upper', 'lower'}
    rule = element.get('rule', None)
    if rule is None:
        if samples is None:
            rule = 'left'
        else:
            rule = 'user-defined'

    if rule is not None:
        if rule == 'left':
            samples = partition[:-1]
        if rule == 'right':
            samples = partition[1:]
        if rule == 'midpoint':
            samples = (partition[:-1] + partition[1:])/2
    try:
        f = un.valid_eval(element.get('function'))
    except:
        log.error(f"Error in <riemann-sum> retrieving function={element.get('function')}")
        return

    annotation = None
    interval_text = None
    if element.get('annotate', 'no') == 'yes':
        annotation = ET.Element('annotation')
        for attrib in ['id', 'text', 'circular', 'sonify', 'speech']:
            if element.get(attrib, None) is not None:
                annotation.set(attrib, element.get(attrib))
        if annotation.get('id', None) is not None:
            annotation.set('ref', annotation.get('id'))
        if annotation.get('text', None) is not None:
            annotation.set('text', label.evaluate_text(annotation.get('text')))
        if annotation.get('speech', None) is not None:
            annotation.set('speech', label.evaluate_text(annotation.get('speech')))
        diagram.push_to_annotation_branch(annotation)
        interval_text = element.get('subinterval-text', None)

    # We will change this element to a group and add area elements below it
    element.tag = 'group'
    outline = element.get('outline', None)
    if outline is None:
        element.set('outline', 'tactile')
    else:
        if outline == 'yes':
            element.set('outline', 'always')
    stroke = element.get('stroke', 'black')
    fill = element.get('fill', 'none')
    thickness = element.get('thickness', '2')
    miterlimit = element.get('miterlimit', None)
    if diagram.output_format() == 'tactile':
        if fill != 'none':
            fill = 'lightgray'

    for interval_num in range(N):
        left = partition[interval_num]
        right = partition[interval_num+1]
        un.enter_namespace('_interval', interval_num)
        un.enter_namespace('_left', f"{left:g}")
        un.enter_namespace('_right', f"{right:g}")
        area = ET.SubElement(element, 'area-under-curve')
        area.set('id', f"{element_id}_{interval_num}")
        area.set('domain', f"({left},{right})")
        area.set('stroke', stroke)
        area.set('fill', fill)
        area.set('thickness', thickness)
        if miterlimit is not None:
            area.set('miterlimit', miterlimit)
        if rule in constant_rules:
            if rule == 'left' or rule == 'right' or rule == 'midpoint' or rule == 'user-defined':
                y_value = f(samples[interval_num])
            if rule == 'upper':
                x_values = np.linspace(left, right, 101)
                y_value = max([f(x) for x in x_values])
            if rule == 'lower':
                x_values = np.linspace(left, right, 101)
                y_value = min([f(x) for x in x_values])
            un.enter_namespace('_height', f"{y_value:g}")
            constant = lambda x, y=y_value: y
            function_name = f"__constant_{interval_num}"
            un.enter_function(function_name, constant)
            area.set('function', function_name)
            area.set('N', '1')
        else:
            if rule == 'trapezoidal':
                area.set('function', element.get('function'))
                area.set('N', '1')
            if rule == 'simpsons':
                h = (right-left)/2
                mid = left+h
                y0 = f(left)
                y1 = f(mid)
                y2 = f(right)
                c = y1
                a = (y0+y2-2*y1)/(2*h*h)
                b = (y2-y0)/(2*h)
                parabola = lambda x, a=a, b=b, c=c, mid=mid: a*(x-mid)**2 + b*(x-mid) + c
                function_name = f"__parabola_{interval_num}"
                un.enter_function(function_name, parabola)
                area.set('function', function_name)
                area.set('N', '100')
        if interval_text is not None:
            interval_annotation = ET.SubElement(annotation, 'annotation')
            interval_annotation.set('ref', area.get('id'))
            interval_annotation.set('text', label.evaluate_text(interval_text))

    group.group(element, diagram, parent, outline_status)
    if annotation is not None:
        diagram.pop_from_annotation_branch()



import lxml.etree as ET
import logging
import math
import numpy as np
import copy
from . import user_namespace as un
from . import utilities as util
from . import math_utilities as math_util
from . import arrow
from . import label

log = logging.getLogger('prefigure')

# Add a graphical element describing a vector
def vector(element, diagram, parent, outline_status):
    if outline_status == 'finish_outline':
        finish_outline(element, diagram, parent)
        return

    # v describes the mathematical vector (displacement),
    # which is scaled by scale, tail is the location of 
    # the tail
    try:
        v = un.valid_eval(element.get('v'))
    except:
        log.error(f"Error parsing vector attribute @v={element.get('v')}")
        return

    tail = un.valid_eval(element.get('tail', '[0,0]'))
    scale = un.valid_eval(element.get('scale', '1'))
    v = scale * v
    w = v + tail

    diagram.register_source_data(element, 'v', v)
    diagram.register_source_data(element, 'head', w)

    # Specify where we want the head to appear.  By default,
    # it's at the tip of the vector, but it can be anywhere
    # along the shaft at a location specified by 0 <= t <= 1
    t = element.get('head-location', None)
    if t is not None:
        t = float(t)
        head_loc = (1-t)*tail + t * w

    if diagram.output_format() == 'tactile':
        element.set('fill', 'black')
        element.set('stroke', 'black')
    else:
        util.set_attr(element, 'stroke', 'black')
        util.set_attr(element, 'fill', 'none')
    util.set_attr(element, 'thickness', '3')

    vector = ET.Element('path')
    diagram.add_id(vector, element.get('id'))
    diagram.register_svg_element(element, vector)
    util.add_attr(vector, util.get_2d_attr(element))

    # Now add the head using an SVG marker
    if t is not None:
        location = 'marker-mid'
    else:
        location = 'marker-end'
    arrow_id = arrow.add_arrowhead_to_path(
        diagram,
        location,
        vector,
        arrow_width=element.get('arrow-width', None),
        arrow_angles=element.get('arrow-angles', None)
    )

    # we need to pull the tip of the vector in a bit to accommodate
    # the arrowhead
    p0 = diagram.transform(tail)
    p1 = diagram.transform(w)
    diff = p1 - p0
    length = math_util.length(diff)
    angle = math.atan2(diff[1], diff[0])
    diagram.register_source_data(element, 'angle', angle)

    arrow_head_length = arrow.get_arrow_length(arrow_id)
    thickness = un.valid_eval(element.get('thickness'))
    if location == 'marker-end':
        length -= thickness * arrow_head_length
        p1 = length * np.array([math.cos(angle),math.sin(angle)]) + p0

    # Here is the shaft of the vector.  If the head is not at
    # the tip, we add a waypoint along the line where the head
    # will appear
    cmds = []

    cmds.append('M ' + util.pt2str(p0))
    if t is not None:
        cmds.append('L ' + util.pt2str(diagram.transform(head_loc)))
    cmds.append('L ' + util.pt2str(p1))
    d = ' '.join(cmds)
    vector.set('d', d)

    if outline_status == 'add_outline':
        diagram.add_outline(element, vector, parent)
        return

    if element.get('outline', 'no') == 'yes' or diagram.output_format() == 'tactile':
        diagram.add_outline(element, vector, parent)
        finish_outline(element, diagram, parent)
    else:
        original_parent = parent
        parent = add_label(element, diagram, parent)
        parent.append(vector)

        # no label has been added if the parent hasn't changed
        if original_parent == parent:
            diagram.register_svg_element(element, vector)
            return

        diagram.register_svg_element(element, parent)
        # if there is a label, then the id is on the outer <g> element
        # so we need to remove it from the children
        if element.get('id', 'none') == parent.get('id'):
            element.attrib.pop('id')
        for child in parent:
            if child.get('id', None) is not None:
                child.attrib.pop('id')

def finish_outline(element, diagram, parent):
    original_parent = parent
    parent = add_label(element, diagram, parent)

    # if we've added a label, remove the id's from element under the parent <g>
    if original_parent != parent:
        for child in parent:
            if child.get('id', None) is not None:
                child.attrib.pop('id')

    diagram.finish_outline(element,
                           element.get('stroke'),
                           element.get('thickness'),
                           element.get('fill', 'none'),
                           parent)

# dictionary for alignments and offsets based on half-quadrant
alignment_dict = {0:('se',-1),
                  1:('nw',1),
                  2:('ne',-1),
                  3:('sw',1),
                  4:('nw',-1),
                  5:('se',1),
                  6:('sw',-1),
                  7:('ne',1)}

def add_label(element, diagram, parent):
    # Is there a label associated with point?
    text = element.text

    # is there a label here?
    has_text = text is not None and len(text.strip()) > 0
    all_comments = all([subel.tag is ET.Comment for subel in element])
    if has_text or not all_comments:
        # If there's a label, we'll bundle the label and vector in a group
        group = ET.SubElement(parent, 'g')
        diagram.add_id(group, element.get('id'))

        # Now we'll create a new XML element describing the label
        el = copy.deepcopy(element)
        el.tag = 'label'

        # determine alignment of the label
        alignment = element.get('alignment', None)
        user_offset = element.get('offset', None)
        angle = diagram.get_source_data(element, 'angle')
        if alignment is None:
            angle_degrees = math.degrees(-angle)
            while angle_degrees < 0:
                angle_degrees += 360
            half_quadrant = math.floor(angle_degrees / 45)
            alignment, offset_dir = alignment_dict[half_quadrant]
            el.set('alignment', alignment)

            normal = (math.cos(-angle), math.sin(-angle))
            direction = offset_dir * np.array((-normal[1], normal[0]))
            offset = 4 * direction
            if user_offset is not None:
                offset += un.valid_eval(user_offset)
            el.set('abs-offset', util.np2str(offset))
        else:
            displacement = label.alignment_displacement[alignment]
            def_offset = np.array([4*(displacement[0]+0.5),
                                   4*(displacement[1]-0.5)])
            if user_offset is not None:
                def_offset += un.valid_eval(user_offset)
            el.set('offset', util.np2str(def_offset))

        head = diagram.get_source_data(element, 'head')
        el.set('anchor', util.pt2long_str(head, spacer=','))

        # add the label graphical element to the group
        label.label(el, diagram, group)
        return group
    else:
        return parent
